//
//  TinkerDevices.h
//
//  Created by Christian Franz on 21/11/2012.
//  Copyright 2012 cf/x AG. All rights reserved.
//     http://www.cfxsoftware.com
//
//  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and 
//  associated documentation files (the "Software"), to deal in the Software without restriction, including 
//  without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or 
//  sell copies of the Software, and to permit persons to whom the Software is furnished to do so, 
//  subject to the following conditions:
//
//  The above copyright notice and this permission notice shall be included in all copies or substantial 
//  portions of the Software and its documentation.
//
//  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT 
//  NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
//  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
//  WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
//  SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//

#import <Cocoa/Cocoa.h>
#include "bricklet_dual_relay.h"
#include "bricklet_io4.h"
#include "bricklet_rotary_poti.h"
#include "bricklet_distance_ir.h"
#include "bricklet_lcd_20x4.h"
#include "bricklet_ambient_light.h"


/*
 
 TinkerDevice is the Root 'Brick/Bricklet' Tinker Device. It is fully NSObject compliant
 and implements most common (and a lot of new) methods for using Bricklets
 
 All TDxxx actual bricks/bricklets descent from this one
 
 */

@interface TinkerDevice : NSObject {
	BOOL isConnected;		// HANDS OFF! TinkerIO changes this - and TinkerIO only!
	BOOL useExceptions; // if true, operations throw exceptions on errors. currently not implemented
	NSString *myUID;	// the UID for this device
	void *myContext;	// user-only: you can set a pointer to anything here. it's always
						// available to this object, even at interrupt time
	id myDelegate;		// my delegate
}

// designated init for all Tinker Devices via Cocoa
- (id) initWithUID :  (const char*) uid;

// general accessors for info shared across all devices, all nicely cocoa-fied
- (Device *) device; // returns my low-level device (e.g. dual_relay), whatever it may be
- (NSString *) deviceType; // what kind of brick this COCOA-Object wraps. 
- (NSString *) deviceName; // returns my device name as NSString.
- (NSString *) uid; // return device UID;
- (NSString *) firmware; // returns the firmware version as NSString
- (NSString *) binding; // returns the binding version as NSString
- (NSString *) description; // returns a string describing brick, firmware, binding. Must be added to connection first! 

// low-level stuff provided for compatibility with existing code
- (void) deviceName : (char [40]) theName; // returns raw device name
- (void) deviceFirmware : (uint8_t [3]) firmWare; // returns raw firmware version
- (void) deviceBinding : (uint8_t [3]) binding; // returns raw binding version


- (void) registerCallback : (void *) callback forID : (uint8_t) theID; // general-purpose callback registration
- (void) registerCallback : (void *) callback;	// bricks/-lets with only one callback automatically use their IRQ ID

- (void) setContext : (void *) data;
- (void *) context;

- (void) setDelegate: (id) aDelegate;
- (id) delegate;

- (bool) connected; // returns true if we have a connection to the device
- (bool) isCompoundDevice; // true for devices that are made from multiple devices (e.g. TIOArray)
- (NSArray *) compoundDevices; // list of all my devices.

@end

/*
 TDProtocol is the delegate protocol that is called during callbacks
 This protocol is entirely provisional, as we first have to see how
 the 2.0 API implements context. Currently not fiunctional.
 
 */
@protocol TDeviceProtocol
- TDCallBackReceived : (TinkerDevice *) device;
- TDDeviceConnected : (TinkerDevice *) device;
- TDDeviceDisconnected : (TinkerDevice *) device;
- TDDeviceWillDeallocate : (TinkerDevice *) device;

@end


//
// TDIODevice is an abstract root device for all combines I/O-Bricklets
// that defines some nice additional features (e.g. valueOfPin:, setPin:toValue:, configurationForPin:input:value: etc.) 
// all other accessors have been harmonized to work with *any* IO bricklet (currently up to 32 pins)
//

@interface TDIODevice : TinkerDevice {
	
}
- (int) numberOfPins;
- (int) setPin : (int) thePin toValue : (bool) val;
- (bool) valueOfPin : (int) thePin;
- (int) setConfigurationForPin : (int) thePin input : (bool) isInput value : (bool) val;
- (int) configurationForPin : (int) thePin input : (bool *) isInput value : (bool *) val;
- (int) setPinConfiguration : (uint32_t) mask forDirection : (char) dir value : (bool) val; 
- (int) pinConfiguration : (uint32_t *) pins andValues : (uint32_t *) vals;

- (int) getPinValues : (uint32_t *) mask; 

- (int) setDebounce : (uint32_t) period;
- (uint32_t) debounce;

- (int) startMonoflopForPin: (int) pin state: (bool) isHigh millisecs: (uint32_t) time; 
- (int) getMonoflopForPin : (uint8_t) pin value : (uint8_t *) val millisecs : (uint32_t *) time remaining : (uint32_t *) timeRemaining;
- (int) setOutputValuesWithMask : (uint32_t) mask; 
- (int) startMonoflopForPinMask : (uint32_t) mask stateMask : (uint32_t) states millisecs : (uint32_t) time;

- (int) setInterruptMask : (uint32_t) mask;
- (uint32_t) interruptMask;
- (int) setInterruptForPin : (int) pin enable : (bool) isHigh; // no mucking about with masks!
- (bool) interruptForPin : (int) pin;

@end


//
// IOArray - holds any number of IO devices as one device
//
@interface TDIOArray : TDIODevice
{
	NSMutableArray *theDevices;
}

- (void) addIOBricklet : (TDIODevice *) bricklet; // add this bricklet to my collection

@end


//
// IO4 Bricklet - 4 IO pins
//
@interface TDIO4 : TDIODevice {
	IO4 io4;
}

@end

//
// IO16 Bricklet - 16 pins
// Behaves exactly like IO4. Note that port 'A' is reagrded as pins 0-7, 'B' as pins 8-15 
//

// will implement as soon as I have one


//
// Abstract Relay Bricklet
// Allows you to transparently switch in your code between 2 and 4 relay bricks
//
@interface TDRelay : TinkerDevice
{
	
}
- (int) numberOfRelays;
- (BOOL) stateOfRelay : (int ) r;
- (void) setRelay : (int) r toState : (BOOL) aState; 
- (void) toggleRelay : (int) r;
- (int) startMonoFlopforRelay : (uint8_t) r toHigh : (BOOL) state forMillisecs : (uint32_t) time; 
- (int) getMonoFlopForRelay : (uint8_t) r isHigh: (BOOL *) retState millisecs : (uint32_t *) time remaining : (uint32_t *) timeRemaining;
	

@end

@interface TDRelayArray : TDRelay
{
	NSMutableArray *theDevices;
}

- (void) addRelayBricklet : (TDRelay *) bricklet;


@end

//
// Dual Relay Bricklet - 2 Relays
//
@interface TDDualRelay : TDRelay {
	DualRelay dr;
	
}
// low-level stuff, only here for compatibility. do not use or you muck up your code
// (i.e. lose all inheritance-enabled switch between brick size advantage)
// use the more general stateOfRelay: and setRelay:toState: instead
- (int) getStateR1 : (BOOL *)r1 R2 : (BOOL *)r2;
- (int) setStateR1 : (BOOL) r1 R2 : (BOOL) r2;


@end

//
// TDSampler Bricklets are all in-only single-channel bricklets that measure something:
// Ambielnt Light, Potis, Baro pressure, ...
//
@interface TDSampler : TinkerDevice
{
	
}
- (int) setDebounce : (uint32_t) period;
- (uint32_t) debounce;

- (int32_t) position; // this is the processed value
- (int) setPositionPeriod : (uint32_t) period;
- (uint32_t) positionPeriod;
- (int) setPositionThreshold : (char) option min : (int16_t) minv max : (int16_t) maxv;
- (int) positionThreshold : (char *) option min : (int16_t *) minv max : (int16_t *) maxv;

- (int32_t) value; // this is the "raw" analog value
- (int) setValuePeriod : (uint32_t) period;
- (uint32_t) valuePeriod;
- (int) setValueThreshold : (char) option min : (int32_t) minv max : (int32_t) maxv;
- (int) valueThreshold : (char *) option min : (int32_t *) minv max : (int32_t *) maxv;


@end

//
// IR Distance bricklet
//
@interface TDIRDistance : TDSampler
{
	DistanceIR dir;
}
// calibration interface
- (int) setSamplingPoint : (uint8_t) point toDistance : (uint16_t) distance;
- (uint16_t) samplingPoint : (uint8_t) point; // returns the distance for this sampling point

@end

//
// Ambient light
//
@interface TDAmbientLight : TDSampler
{
	AmbientLight ambi;
}


@end


//
// TDPotentiometer is the superclass to all potis: linear and rotary
//
@interface TDPotentiometer : TDSampler
{
}


@end

//
// Rotary Poti
//
@interface TDRotaryPot : TDPotentiometer{
	RotaryPoti rotP;	

}

@end

//
// TDLCD - Bricklets with an LCD screen
// This virtual class implements an LCD screen so you can 
// writeln and goto in your code, and have a mirror of
// what is on-screen. If you enable lcdSync, all changes
// are written immediately. Otherwise the LCD is only
// updated when you call updateLCD
//
#define cLCDMaxChar 20 // currrently no LCD display has more than 20 chars. chnage if we get bigger ones
#define cLCDMaxLine 4; 
@interface TDLCD : TinkerDevice {
	// the following is the pseudo-lcd text buffer
	int cx, cy; // cursor pos
	int maxX, maxY; // dimensions of LCD *DEVICE*
	char l1[cLCDMaxChar + 1]; // 21st is always 0x0 for easy c-string conversion
	char l2[cLCDMaxChar + 1];
	char l4[cLCDMaxChar + 1];
	char l3[cLCDMaxChar + 1];

	bool useAutosync; // if true, changes are written immediately 
}

- (void) clearScreen;	// set whole screen to 0x20
- (void) clearEOL;		// from current cursor pos to EOL
- (void) scrollLine;	// scrolls all lines up, fills line 4 with 0x20
- (char *) currline;	// returns address of current line, first chat
- (void) writeChar : (char) theChar; // write a single char at current cursor pos, advance pos
- (void) writeString : (NSString *) aString; // write at x,y no CR
- (void) writeLnString : (NSString *) aString; // write, CR, clear EOL, scroll if necessary
- (void) moveToX : (int) xp Y: (int) yp; // set cursor position
- (void) home; // clear screen, move to 0,0;
- (void) cursorX : (int *) x Y : (int *) y; 
- (void) logAll; // dump contents of 'lcd' buffer to log. for debugging purposes

- (void) setAutosync : (bool) sync;
- (bool) autosync;
- (void) updateLCD; // copies lcd buffer to device

// device convenience
- (bool) backlight;
- (void) setBacklight : (bool) on;
- (bool) button: (int) button;
- (void) setCursor: (bool) on blinking : (bool) blink;
- (void) getCursor: (bool *) on blinking : (bool *) blink;

// low-level for compatibility
- (void) backlightOn;
- (void) backlightOff;
- (void) clearLCD;
- (void) writeLine : (char *) text atX : (uint8_t) x Y : (uint8_t) y;

@end

@interface TDLCD20x4 : TDLCD
{
	LCD20x4 theLCD;
}


@end




