//
//  TinkerDevices.m
//
//  Created by Christian Franz on 21/11/2012.
//  Copyright 2012 cf/x AG. All rights reserved.
//     http://www.cfxsoftware.com
//
//  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and 
//  associated documentation files (the "Software"), to deal in the Software without restriction, including 
//  without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or 
//  sell copies of the Software, and to permit persons to whom the Software is furnished to do so, 
//  subject to the following conditions:
//
//  The above copyright notice and this permission notice shall be included in all copies or substantial 
//  portions of the Software and its documentation.
//
//  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT 
//  NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
//  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
//  WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
//  SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//

#import "TinkerDevices.h"

#pragma mark TinkerDevice ROOT CLASS
//
// Abstract Tinker Device
//
@implementation TinkerDevice
- initWithUID: (const char*) uid {
	self = [super init];
	if (self) {
		myUID = [NSString stringWithCString:uid
								   encoding:NSMacOSRomanStringEncoding];
		[myUID retain];
	}
	return (self); // should never be called
}

- (Device *) device {
	return nil;
}

- (NSString *) deviceName { // returns my device name as NSString.
	char theName[40];
	[self deviceName:theName]; // WARNING: note the similary named methods 'deviceName' and 'deviceName:'
	return [NSString stringWithCString:theName 
							  encoding:NSMacOSRomanStringEncoding];
}	

- (NSString *) uid {
	return myUID;
}

- (NSString *) firmware {
	uint8_t fw[3];
	[self deviceFirmware:fw];
	return [NSString stringWithFormat:@"%d.%d.%d", fw[0], fw[1], fw[2] ];
	
}

- (NSString *) binding {
	uint8_t bind[3];
	[self deviceBinding:bind];
	return [NSString stringWithFormat:@"%d.%d.%d", bind[0], bind[1], bind[2] ];
	
}

- (NSString *) description {
	if ([self connected]) {
		return [NSString stringWithFormat:@"%@-Bricklet: %@ with UID = %@: Firmware = %@, Binding = %@", [self deviceType], [self deviceName], [self uid], [self firmware], [self binding]];
	} else {
		return [NSString stringWithFormat:@"Bricklet type %@ (UID = %@) not connected or missing in stack", [self deviceType], [self uid]];
	};

}
- (NSString *) deviceType {
	return @"Virtual TinkerDevice";
}

- (void) deviceName : (char [40]) theName {
	theName[0] = 0;
}

- (void) deviceFirmware : (uint8_t [3]) firmWare {
	firmWare[0] = 0; firmWare [1] = 0; firmWare[2] = 0;
}

- (void) deviceBinding : (uint8_t [3]) binding {
	binding[0] = 0; binding[1] = 0; binding[2] = 0; 
}

- (void) registerCallback : (void *) callback forID : (uint8_t) theID {
}

- (void) registerCallback : (void *) callback {
}

- (void) setContext : (void *) data {myContext = data;}
- (void *) context {return myContext;}

-(void) setDelegate:(id)aDelegate { myDelegate = aDelegate;};
- (id) delegate {return myDelegate;}

- (bool) connected {return isConnected;}

- (bool) isCompoundDevice {return false;}
- (NSArray *) compoundDevices{return [NSArray arrayWithObject:self];}

@end


#pragma mark BRICKLETS

#pragma mark IO Superclass
// 
// abstract IO class, so we can transparently switch between 4 and 16 IO blocks
//
@implementation TDIODevice

- (int) setPin : (int) thePin toValue : (bool) val {
	// create mask 
	uint32_t mask = 1 << thePin;
	uint32_t iMask = ~mask;
	uint32_t shiftedVal = 0;
	if (val) { shiftedVal = 1 << thePin; };
	
	[self getPinValues: &mask];
	mask = mask & iMask;
	mask = mask | shiftedVal;
	
	return [self setOutputValuesWithMask : mask];
}	

- (int) numberOfPins {return 0;}

- (bool) valueOfPin : (int) thePin {
	// create mask
	uint32_t mask; 
	uint32_t imask = 1 << thePin;
	[self getPinValues: &mask];
	
	if ( mask & imask) { return true;} else {return false;};
}

- (int) setOutputValuesWithMask : (uint32_t) mask {return -1;} 
- (int) getPinValues : (uint32_t *) mask { *mask = 0; return -1;}
- (int) setPinConfiguration : (uint32_t) mask forDirection : (char) dir value : (bool) val {return -1;}
- (int) pinConfiguration : (uint32_t *) pins andValues : (uint32_t *) vals {return -1;}

- (int) setConfigurationForPin : (int) thePin input : (bool) isInput value : (bool) val {
	char dir = 'i';
	if (!isInput) {
		dir = 'o';
	};
	
	uint32_t mask = 1 << thePin;
	return [self setPinConfiguration:mask 
						forDirection:dir 
							   value:val];
	
}

- (int) configurationForPin : (int) thePin input : (bool *) isInput value : (bool *) val {
	uint32_t mask = 1 << thePin;
	uint32_t dirs = 0;
	uint32_t vals = 0;
	int res = [self pinConfiguration:&dirs 
						   andValues:&vals];
	*isInput = false;
	*val = false;
	if (dirs & mask) {*isInput = true;};
	if (vals & mask) {*val = true;};
	return res;
}

- (int) setDebounce : (uint32_t) period {return -1;}
- (uint32_t) debounce {return 0;}
- (int) startMonoflopForPinMask : (uint32_t) mask stateMask : (uint32_t) states millisecs : (uint32_t) time {return -1;}

- (int) startMonoflopForPin: (int) pin state: (bool) isHigh millisecs: (uint32_t) time {
	uint32_t mask = 1 << pin;
	uint32_t stateMask = 0;
	if (isHigh) {stateMask = mask;};
	return [self startMonoflopForPinMask:mask 
							   stateMask:stateMask 
							   millisecs:time];
}

- (int) getMonoflopForPin : (uint8_t) pin value : (uint8_t *) val millisecs : (uint32_t *) time remaining : (uint32_t *) timeRemaining {return -1;}
- (int) setInterruptMask : (uint32_t) mask {return -1;}
- (uint32_t) interruptMask {return 0;}

- (int) setInterruptForPin : (int) pin enable : (bool) isHigh;{ // no mucking about with masks!
	// all examples for pin 2
	uint32_t mask = 1 << pin; // 0100  0100
	uint32_t eraser = ~mask;  // 1011  1011
	if (!isHigh) {mask = 0;};  // 0100  0000
	mask = mask | eraser;	  // 1111  1011
	
	uint32_t currMask = [self interruptMask];
	// make sure bit pin is on
	currMask = currMask | (1 << pin);
	currMask = currMask & mask; // now it is off if isHigh was false
	
	return	[self setInterruptMask:currMask];
}

- (bool) interruptForPin : (int) pin {
	uint32_t mask = [self interruptMask];
	if (mask & (1<<pin)) {return TRUE;} else {return false;}; 
}

@end

#pragma mark IOArray
//
// IOArray
//
@implementation TDIOArray
- initWithUID: (const char*) uid {
	self = [super init];
	if (self) {
		theDevices = [[NSMutableArray alloc] initWithCapacity:10];
	}
	return (self); // should never be called
}

- (NSString *) deviceType {
	return @"IO Array";
}

- (void) deviceName : (char [40]) theName {
	theName = "Cocoa Layer IO Array";
}

- (void) deviceFirmware : (uint8_t [3]) firmWare {
	firmWare[0] = 1; firmWare [1] = 0; firmWare[2] = 0;
}

- (void) deviceBinding : (uint8_t [3]) binding {
	binding[0] = 1; binding[1] = 0; binding[2] = 0; 
}

- (bool) connected {
	bool allIsWell = TRUE;
	TinkerDevice *aDevice = nil;
	for (aDevice in theDevices) {
		allIsWell = allIsWell & [aDevice connected];
	}
	
	return allIsWell;
}

- (int) numberOfPins {
	int sum = 0;
	TDIODevice *aDevice = nil;
	for (aDevice in theDevices) {
		sum = sum + [aDevice numberOfPins];
	};
	
	return sum;
	
}

- (TDIODevice *) device4pin: (int) thePin newPin : (int *) modPin {
	// get the device to which thePin belongs
	*modPin = 0;
	if (thePin >= [self numberOfPins]) {return nil;}
	
	TDIODevice *aDevice = nil;
	uint32 index = 0;
		
	while (index < [theDevices count]) {
		aDevice = [theDevices objectAtIndex:index];
		if (thePin < [aDevice numberOfPins]) {
			// now thePin is in range of this device
			*modPin = thePin; // this is the requested pin #
			return aDevice; // return this device
		};
		// nope. subtract the number of pins for this device, iterate
		thePin = thePin - [aDevice numberOfPins];
		index = index + 1;
	} 
			
	return nil;
}

- (bool) valueOfPin : (int) thePin {
	int newPin;
	TDIODevice *newDevice;
	newDevice = [self device4pin:thePin newPin:&newPin];
	return [newDevice valueOfPin:newPin];
}

- (int) setPin : (int) thePin toValue : (bool) val {
	int newPin;
	TDIODevice *newDevice;
	newDevice = [self device4pin:thePin newPin:&newPin];
	return [newDevice setPin:newPin toValue:val];
}

- (int) setConfigurationForPin : (int) thePin input : (bool) isInput value : (bool) val {
	int newPin;
	TDIODevice *newDevice;
	newDevice = [self device4pin:thePin newPin:&newPin];
	return [newDevice setConfigurationForPin:newPin input:isInput value:val];
}

- (int) configurationForPin : (int) thePin input : (bool *) isInput value : (bool *) val {
	int newPin;
	TDIODevice *newDevice;
	newDevice = [self device4pin:thePin newPin:&newPin];
	return [newDevice configurationForPin:newPin input: isInput value:val];
}

- (int) setDebounce : (uint32_t) period {
	TDIODevice *aDevice = nil;
	for (aDevice in theDevices) {
		return [aDevice setDebounce:period];
	};
	
	return 0;
}

- (uint32_t) debounce {
	// simply return the debounce of the first device
	if ([theDevices count] < 1) {return 0;};
	return [[theDevices objectAtIndex:0] debounce];
}

- (int) startMonoflopForPin: (int) pin state: (bool) isHigh millisecs: (uint32_t) time {
	int newPin;
	TDIODevice *newDevice;
	newDevice = [self device4pin:pin newPin:&newPin];
	return [newDevice startMonoflopForPin:newPin state:isHigh millisecs:time];
}

- (int) getMonoflopForPin : (uint8_t) pin value : (uint8_t *) val millisecs : (uint32_t *) time remaining : (uint32_t *) timeRemaining {
	int newPin;
	TDIODevice *newDevice;
	newDevice = [self device4pin:pin newPin:&newPin];
	return [newDevice getMonoflopForPin:newPin value:val millisecs:time remaining:timeRemaining];
	
}

- (int) setInterruptForPin : (int) pin enable : (bool) isHigh;{ // no mucking about with masks!
	int newPin;
	TDIODevice *newDevice;
	newDevice = [self device4pin:pin newPin:&newPin];
	return [newDevice setInterruptForPin:newPin enable:isHigh];
	
}

- (bool) interruptForPin : (int) pin {
	int newPin;
	TDIODevice *newDevice;
	newDevice = [self device4pin:pin newPin:&newPin];
	return [newDevice interruptForPin:newPin];
}

- (void) addIOBricklet:(TDIODevice *)bricklet {
	// first, set debounce to common debounce val
	if ([theDevices count] > 0) {
		[bricklet setDebounce:[[theDevices objectAtIndex:0] debounce]];
	};
	
	[theDevices addObject:bricklet];
}

- (bool) isCompoundDevice {return true;} // true for devices that are made from multiple devices (e.g. TIOArray)
- (NSArray *) compoundDevices{return theDevices;} // list of all my devices.

@end


#pragma mark IO4
//
// IO4
//
@implementation TDIO4
-(id) initWithUID: (const char *) uid {
	self = [super initWithUID:uid];
	if (self) {
		// create the bricklet
		io4_create(&io4, uid); 
		
	}
	return (self);
}

- (Device *) device {
	return &io4;
}

- (int) numberOfPins {return 4;}

- (NSString *) deviceType {
	return @"IO4";
}

- (int) setOutputValuesWithMask : (uint32_t) mask {
	uint8_t smallmask = mask;
	return io4_set_value([self device], smallmask);
}

- (int) getPinValues : (uint32_t *) mask {
	
	uint8_t smallmask;
	*mask = 0;
	int res = io4_get_value([self device], &smallmask);
	*mask = *mask | smallmask;
	return res;
}


- (int) setPinConfiguration : (uint32_t) mask forDirection : (char) dir value : (bool) val {
	uint8_t smallmask = mask;
	return io4_set_configuration([self device], smallmask, dir, val);
	
}

- (int) pinConfiguration : (uint32_t *) pins andValues : (uint32_t *) vals {
	uint8_t dirs = 0;
	uint8_t values = 0;
	int res = io4_get_configuration( [self device], &dirs, &values);
	*pins = 0 | dirs;
	*vals = 0 | values;
	return res;
}



- (int) setDebounce : (uint32_t) period {
	return io4_set_debounce_period([self device], period);
}

- (uint32_t) debounce {//: (uint32_t *) period {
	uint32_t period = 0;
	io4_get_debounce_period([self device], &period); 
	return period;
}

- (int) startMonoflopForPinMask : (uint32_t) mask stateMask : (uint32_t) states millisecs : (uint32_t) time {
	uint8_t smallmask = mask;
	uint8_t	smallstate = states;
	return io4_set_monoflop( [self device], smallmask, smallstate, time);
}


- (int) getMonoflopForPin : (uint8_t) pin value : (uint8_t *) val millisecs : (uint32_t *) time remaining : (uint32_t *) timeRemaining {
	return io4_get_monoflop([self device], pin, val, time, timeRemaining);
}


- (int) setInterruptMask : (uint32_t) mask {
	uint8_t smallmask = mask;
	return io4_set_interrupt([self device], smallmask);
}

- (uint32_t) interruptMask {
	uint8_t smallmask = 0;
	uint32_t mask = 0;
	io4_get_interrupt([self device], &smallmask);
	mask = 0 | smallmask;
	return mask;
}


- (void) deviceName : (char [40]) theName { // must be at least 40 chars wide
	uint8_t firmware[3];
	uint8_t binding[3];
	
	io4_get_version([self device], theName, firmware, binding);
}

- (void) deviceFirmware : (uint8_t [3]) firmware { 
	char theName[40];
	uint8_t binding[3];
	
	io4_get_version([self device], theName, firmware, binding);
}


- (void) deviceBinding : (uint8_t [3]) binding { 
	char theName[40];
	uint8_t firmware[3];
	
	io4_get_version([self device], theName, firmware, binding);
}

- (void) registerCallback : (void *) callback forID : (uint8_t) theID {
	io4_register_callback([self device], theID, callback);
}

@end

#pragma mark Relay Superclass
//
// Abstract Relay, for 2 and 4 bricklets
//
@implementation TDRelay
- (int) numberOfRelays {return 0;}
- (BOOL) stateOfRelay : (int ) r {return false;}
- (void) setRelay : (int) r toState : (BOOL) aState {}; 

- (void) toggleRelay:(int) r {
	bool theState = [self stateOfRelay:r];
	[self setRelay:r toState: (!theState)];
	
}
- (int) startMonoFlopforRelay : (uint8_t) r toHigh : (BOOL) state forMillisecs : (uint32_t) time {return -1;} 
- (int) getMonoFlopForRelay : (uint8_t) r isHigh: (BOOL *) retState millisecs : (uint32_t *) time remaining : (uint32_t *) timeRemaining {return 0;}

@end


#pragma mark Relay Array
//
// Relay Array
//
@implementation TDRelayArray
- (NSString *) deviceType {
	return @"IO Array";
}

- (void) deviceName : (char [40]) theName {
	theName = "Cocoa Layer Relay Array";
}

- (void) deviceFirmware : (uint8_t [3]) firmWare {
	firmWare[0] = 1; firmWare [1] = 0; firmWare[2] = 0;
}

- (void) deviceBinding : (uint8_t [3]) binding {
	binding[0] = 1; binding[1] = 0; binding[2] = 0; 
}

- (bool) connected {
	bool allIsWell = TRUE;
	TinkerDevice *aDevice = nil;
	for (aDevice in theDevices) {
		allIsWell = allIsWell & [aDevice connected];
	}
	
	return allIsWell;
}

- (int) numberOfRelays {
	int sum = 0;
	TDRelayArray *aDevice = nil;
	for (aDevice in theDevices) {
		sum = sum + [aDevice numberOfRelays];
	};
	
	return sum;
	
}

- (TDRelay *) device4relay: (int) theRelay newRelay : (int *) modR {
	// get the device to which thePin belongs
	*modR = 0;
	if (theRelay < 1) {return nil;};
	if (theRelay > [self numberOfRelays]) {return nil;};
	
	TDRelay *aDevice = nil;
	uint32 index = 0;
	
	while (index < [theDevices count]) {
		aDevice = [theDevices objectAtIndex:index];
		if (theRelay <= [aDevice numberOfRelays]) {
			// now thePin is in range of this device
			*modR = theRelay; // this is the requested pin #
			return aDevice; // return this device
		};
		// nope. subtract the number of pins for this device, iterate
		theRelay = theRelay - [aDevice numberOfRelays];
		index = index + 1;
	} 
	
	return nil;
}

- (BOOL) stateOfRelay : (int) r {
	int newR;
	TDRelay *newDevice;
	newDevice = [self device4relay:r newRelay:&newR];
	return [newDevice stateOfRelay:newR];
}

- (void) setRelay : (int) r toState : (BOOL) aState {
	int newR;
	TDRelay *newDevice;
	newDevice = [self device4relay:r newRelay:&newR];
	return [newDevice setRelay:newR toState:aState];
}

- (void) toggleRelay:(int)r {
	int newR;
	TDRelay *newDevice;
	newDevice = [self device4relay:r newRelay:&newR];
	[newDevice toggleRelay:newR];
}


- (int) startMonoFlopforRelay: (uint8_t) r toHigh: (BOOL) state forMillisecs: (uint32_t) time {
	int newR;
	TDRelay *newDevice;
	newDevice = [self device4relay:r newRelay:&newR];
	return [newDevice startMonoFlopforRelay:newR toHigh:state forMillisecs:time];
}

- (int) getMonoFlopForRelay : (uint8_t) r isHigh: (BOOL *) retState millisecs : (uint32_t *) time remaining : (uint32_t *) timeRemaining {
	int newR;
	TDRelay *newDevice;
	newDevice = [self device4relay:r newRelay:&newR];
	return [newDevice getMonoFlopForRelay:newR isHigh:retState millisecs:time remaining:timeRemaining];
	
}


- (void) addRelayBricklet :(TDRelay *) bricklet {
	
	[theDevices addObject:bricklet];
}

- (bool) isCompoundDevice {return true;} // true for devices that are made from multiple devices (e.g. TIOArray)
- (NSArray *) compoundDevices{return theDevices;} // list of all my devices.



@end


#pragma mark Dual Relay
//
// Dual Relay
//
@implementation TDDualRelay
-(id) initWithUID: (const char *) uid {
	self = [super initWithUID:uid];
	if (self) {
		// create the bricklet
		dual_relay_create(&dr, uid); 
		
	}
	return (self);
}

- (int) numberOfrelays {return 2;}

- (Device *) device { 
	return &dr; 
}

- (NSString *) deviceType {
	return @"Dual Relay";
}

- (int) getStateR1 : (BOOL *)r1 R2 : (BOOL *)r2 {
	int res = 0;
	res = dual_relay_get_state([self device], r1, r2);
	return res;
}

- (BOOL) stateOfRelay : (int ) r {
	BOOL r1, r2;
	dual_relay_get_state([self device], &r1, &r2);
	if (r == 1) { return r1;} else {return r2;}
}

- (int) setStateR1 : (BOOL) r1 R2 : (BOOL) r2 {
	return dual_relay_set_state( [self device], r1, r2);
}

- (void) setRelay : (int) r toState : (BOOL) aState {
	BOOL last1, last2;
	[self getStateR1:&last1 R2:&last2];
	if (r == 1) { 
		last1 = aState;
	} else {
		last2 = aState;
	};
	[self setStateR1:last1 R2:last2];
}


- (int) startMonoFlopforRelay : (uint8_t) r toHigh : (BOOL) state forMillisecs : (uint32_t) time {
	return dual_relay_set_monoflop([self device], r, state, time);
}

- (int) getMonoFlopForRelay : (uint8_t) r isHigh: (BOOL *) retState millisecs : (uint32_t *) time remaining : (uint32_t *) timeRemaining {

	return dual_relay_get_monoflop([self device], r, retState, time, timeRemaining);
}

- (void) deviceName : (char [40]) theName { // must be at least 40 chars wide
	uint8_t firmware[3];
	uint8_t binding[3];
	
	dual_relay_get_version([self device], theName, firmware, binding);
}

- (void) deviceFirmware : (uint8_t [3]) firmware { 
	char theName[40];
	uint8_t binding[3];
	
	dual_relay_get_version([self device], theName, firmware, binding);
}


- (void) deviceBinding : (uint8_t [3]) binding { 
	char theName[40];
	uint8_t firmware[3];
	
	dual_relay_get_version([self device], theName, firmware, binding);
}

- (void) registerCallback : (void *) callback forID : (uint8_t) theID {
	dual_relay_register_callback([self device], theID, callback);
}

- (void) registerCallback:(void *)callback {
	[self registerCallback:callback 
					 forID:DUAL_RELAY_CALLBACK_MONOFLOP_DONE];
}
@end

#pragma mark Sampler Superclass
@implementation TDSampler
- (int) setDebounce : (uint32_t) period {return -1;}
- (uint32_t) debounce {return 0;}
- (int32_t) position {	return 0;}
- (int) setPositionPeriod : (uint32_t) period {return -1;}
- (uint32_t) positionPeriod { return 0;}
- (int) setPositionThreshold : (char) option min : (int16_t) minv max : (int16_t) maxv {return -1;}
- (int) positionThreshold : (char *) option min : (int16_t *) minv max : (int16_t *) maxv { return -1;}
- (int32_t) value{ return 0;}
- (int) setValuePeriod : (uint32_t) period { return -1;}
- (uint32_t) valuePeriod {return 0; }
- (int) setValueThreshold : (char) option min : (int32_t) minv max : (int32_t) maxv { return -1;}
- (int) valueThreshold : (char *) option min : (int32_t *) minv max : (int32_t *) maxv { return -1;}



@end


#pragma mark IR Distance Sensor
@implementation TDIRDistance
-(id) initWithUID: (const char *) uid {
	self = [super initWithUID:uid];
	if (self) {
		// create the bricklet
		distance_ir_create(&dir, uid); 
		
	}
	return (self);
}

- (Device *) device { 
	return &dir; 
}

- (NSString *) deviceType {
	return @"IR Distance";
}

- (void) deviceName : (char [40]) theName { // must be at least 40 chars wide
	uint8_t firmware[3];
	uint8_t binding[3];
	
	distance_ir_get_version([self device], theName, firmware, binding);
}

- (void) deviceFirmware : (uint8_t [3]) firmware { 
	char theName[40];
	uint8_t binding[3];
	
	distance_ir_get_version([self device], theName, firmware, binding);
}


- (void) deviceBinding : (uint8_t [3]) binding { 
	char theName[40];
	uint8_t firmware[3];
	
	distance_ir_get_version([self device], theName, firmware, binding);
}

- (void) registerCallback : (void *) callback forID : (uint8_t) theID {
	distance_ir_register_callback([self device], theID, callback);
}

- (int) setDebounce : (uint32_t) period {
	return distance_ir_set_debounce_period([self device], period);
}

- (uint32_t) debounce {//: (uint32_t *) period {
	uint32_t period = 0;
	distance_ir_get_debounce_period([self device], &period); 
	return period;
}

- (int32_t) position {
	uint16_t pos;
	distance_ir_get_distance([self device], &pos);
	return pos;
}

- (int) setPositionPeriod : (uint32_t) period {
	return distance_ir_set_distance_callback_period([self device], period);
}

- (uint32_t) positionPeriod {
	uint32_t period;
	distance_ir_get_distance_callback_period([self device], &period);
	return period;
}

- (int) setPositionThreshold : (char) option min : (int16_t) minv max : (int16_t) maxv {
	return	distance_ir_set_distance_callback_threshold([self device], option, minv, maxv);
}

- (int) positionThreshold : (char *) option min : (int16_t *) minv max : (int16_t *) maxv {
	return distance_ir_get_distance_callback_threshold([self device], option, minv, maxv);
}


- (int32_t) value{
	uint16_t val;
	distance_ir_get_analog_value([self device], &val);
	return val;
}

- (int) setValuePeriod : (uint32_t) period {
	return distance_ir_set_analog_value_callback_period([self device], period);
}

- (uint32_t) valuePeriod {
	uint32_t period;
	distance_ir_get_analog_value_callback_period([self device], &period);
	return period;
}

- (int) setValueThreshold : (char) option min : (int32_t) minv max : (int32_t) maxv{
	uint16_t min = minv;
	uint16_t max = maxv;
	return	distance_ir_set_analog_value_callback_threshold([self device], option, min, max);
}


- (int) valueThreshold : (char *) option min : (int32_t *) minv max : (int32_t *) maxv{
	uint16_t min;
	uint16_t max;
	return distance_ir_get_analog_value_callback_threshold([self device], option, &min, &max);
	*minv = min;
	*maxv = max;
}

// calibration interface
- (int) setSamplingPoint : (uint8_t) point toDistance : (uint16_t) distance {
	return distance_ir_set_sampling_point([self device], point, distance);
}

- (uint16_t) samplingPoint : (uint8_t) point { // returns the distance for this sampling point
	uint16_t dist = 0;
	distance_ir_get_sampling_point([self device], point, &dist);
	return dist;
}

@end

#pragma mark Ambient Light Sensor
//
// Ambient Light Sensor
//
@implementation TDAmbientLight



-(id) initWithUID: (const char *) uid {
	self = [super initWithUID:uid];
	if (self) {
		// create the bricklet
		ambient_light_create(&ambi, uid); 
		
	}
	return (self);
}

- (Device *) device { 
	return &ambi; 
}

- (NSString *) deviceType {
	return @"Ambient Light";
}

- (void) deviceName : (char [40]) theName { // must be at least 40 chars wide
	uint8_t firmware[3];
	uint8_t binding[3];
	
	ambient_light_get_version([self device], theName, firmware, binding);
}

- (void) deviceFirmware : (uint8_t [3]) firmware { 
	char theName[40];
	uint8_t binding[3];
	
	ambient_light_get_version([self device], theName, firmware, binding);
}


- (void) deviceBinding : (uint8_t [3]) binding { 
	char theName[40];
	uint8_t firmware[3];
	
	ambient_light_get_version([self device], theName, firmware, binding);
}

- (void) registerCallback : (void *) callback forID : (uint8_t) theID {
	ambient_light_register_callback([self device], theID, callback);
}

- (int) setDebounce : (uint32_t) period {
	return ambient_light_set_debounce_period([self device], period);
}

- (uint32_t) debounce {//: (uint32_t *) period {
	uint32_t period = 0;
	ambient_light_get_debounce_period([self device], &period); 
	return period;
}

- (int32_t) position {
	uint16_t pos;
	ambient_light_get_illuminance([self device], &pos);
	return pos;
}

- (int) setPositionPeriod : (uint32_t) period {
	return ambient_light_set_illuminance_callback_period([self device], period);
}

- (uint32_t) positionPeriod {
	uint32_t period;
	ambient_light_get_illuminance_callback_period([self device], &period);
	return period;
}

- (int) setPositionThreshold : (char) option min : (int16_t) minv max : (int16_t) maxv {
	return	ambient_light_set_illuminance_callback_threshold([self device], option, minv, maxv);
}

- (int) positionThreshold : (char *) option min : (int16_t *) minv max : (int16_t *) maxv {
	return ambient_light_get_illuminance_callback_threshold([self device], option, minv, maxv);
}


- (int32_t) value{
	uint16_t val;
	ambient_light_get_analog_value([self device], &val);
	return val;
}

- (int) setValuePeriod : (uint32_t) period {
	return ambient_light_set_analog_value_callback_period([self device], period);
}

- (uint32_t) valuePeriod {
	uint32_t period;
	ambient_light_get_analog_value_callback_period([self device], &period);
	return period;
}

- (int) setValueThreshold : (char) option min : (int32_t) minv max : (int32_t) maxv{
	uint16_t min = minv;
	uint16_t max = maxv;
	return	ambient_light_set_analog_value_callback_threshold([self device], option, min, max);
}


- (int) valueThreshold : (char *) option min : (int32_t *) minv max : (int32_t *) maxv{
	uint16_t min;
	uint16_t max;
	return ambient_light_get_analog_value_callback_threshold([self device], option, &min, &max);
	*minv = min;
	*maxv = max;
}

// calibration interface
- (int) setSamplingPoint : (uint8_t) point toDistance : (uint16_t) distance {
	return distance_ir_set_sampling_point([self device], point, distance);
}

@end


#pragma mark Potentiometer Superclass
//
// All potentiometers
//
@implementation TDPotentiometer

@end



#pragma mark Rotary Pot
//
// Rotary Pot
//
@implementation TDRotaryPot

-(id) initWithUID: (const char *) uid {
	self = [super initWithUID:uid];
	if (self) {
		// create the bricklet
		rotary_poti_create(&rotP, uid); 
		
	}
	return (self);
}

- (Device *) device { 
	return &rotP; 
}

- (NSString *) deviceType {
	return @"Rotary Potentiometer";
}

- (void) deviceName : (char [40]) theName { // must be at least 40 chars wide
	uint8_t firmware[3];
	uint8_t binding[3];
	
	rotary_poti_get_version([self device], theName, firmware, binding);
}

- (void) deviceFirmware : (uint8_t [3]) firmware { 
	char theName[40];
	uint8_t binding[3];
	
	rotary_poti_get_version([self device], theName, firmware, binding);
}


- (void) deviceBinding : (uint8_t [3]) binding { 
	char theName[40];
	uint8_t firmware[3];
	
	rotary_poti_get_version([self device], theName, firmware, binding);
}

- (void) registerCallback : (void *) callback forID : (uint8_t) theID {
	rotary_poti_register_callback([self device], theID, callback);
}

- (int) setDebounce : (uint32_t) period {
	return rotary_poti_set_debounce_period([self device], period);
}

- (uint32_t) debounce {//: (uint32_t *) period {
	uint32_t period = 0;
	rotary_poti_get_debounce_period([self device], &period); 
	return period;
}

- (int32_t) position {
	int16_t pos;
	rotary_poti_get_position([self device], &pos);
	return pos;
}

- (int) setPositionPeriod : (uint32_t) period {
	return rotary_poti_set_position_callback_period([self device], period);
}

- (uint32_t) positionPeriod {
	uint32_t period;
	rotary_poti_get_position_callback_period([self device], &period);
	return period;
}

- (int) setPositionThreshold : (char) option min : (int16_t) minv max : (int16_t) maxv {
	return	rotary_poti_set_position_callback_threshold([self device], option, minv, maxv);
}

- (int) positionThreshold : (char *) option min : (int16_t *) minv max : (int16_t *) maxv {
	return rotary_poti_get_position_callback_threshold([self device], option, minv, maxv);
}


- (int32_t) value{
	uint16_t val;
	rotary_poti_get_analog_value([self device], &val);
	return val;
}

- (int) setValuePeriod : (uint32_t) period {
	return rotary_poti_set_analog_value_callback_period([self device], period);
}

- (uint32_t) valuePeriod {
	uint32_t period;
	rotary_poti_get_analog_value_callback_period([self device], &period);
	return period;
}

- (int) setValueThreshold : (char) option min : (int32_t) minv max : (int32_t) maxv{
	uint16_t min = minv;
	uint16_t max = maxv;
	return	rotary_poti_set_analog_value_callback_threshold([self device], option, min, max);
}


- (int) valueThreshold : (char *) option min : (int32_t *) minv max : (int32_t *) maxv{
	uint16_t min;
	uint16_t max;
	return rotary_poti_get_analog_value_callback_threshold([self device], option, &min, &max);
	*minv = min;
	*maxv = max;
}

@end


#pragma mark LCD Devices Superclass

@implementation TDLCD

-(id) initWithUID: (const char *) uid {
	self = [super initWithUID:uid];
	if (self) {
		// this virtual device has cMaxChar chars and 4 lines
		maxX = cLCDMaxChar;
		maxY = cLCDMaxLine;
		useAutosync = true;
	}
	return (self);
}

- (char *) currline {
	char *currline = nil;
	if (cy == 0) {currline = l1;} 
	else if (cy == 1) {currline = l2;} 
	else if (cy == 2) {currline = l3;} 
	else {currline = l4;};
	
	return currline;
}

-(void) clearScreen {
	// set all chars to 0x20 (space)
	int i = 0;
	while (i < maxX) {
		l1[i] = 0x20;
		l2[i] = 0x20;
		l3[i] = 0x20;
		l4[i] = 0x20;
		
		i = i + 1;
	}
	l1[maxX] = 0; // make compatible with c-string
	l2[maxX] = 0;
	l3[maxX] = 0;
	l4[maxX] = 0;

	if ([self autosync]) {[self updateLCD];}; 

}

- (void) clearEOL {
	char *currLine = [self currline];
	int savex = cx;
	while (cx < maxX) {
		currLine[cx] = 0x20;
		cx = cx + 1;
	}
	cx = savex;

	if ([self autosync]) {[self updateLCD];}; 

}

- (void) scrollLine {
	// scroll all up one, last line:
	// set all chars to 20 (space)
	// we assume to have at least 2 lines of display
	int i = 0;
	while (i < maxX) {
		l1[i] = l2[i]; // copy 2 -> 1
		if (maxY > 2) {l2[i] = l3[i];} else {l2[i] = 0x20;};
		if (maxY > 3) {l3[i] = l4[i];} else {l3[i] = 0x20;};
		l4[i] = 0x20;
		
		i = i + 1;
	}
	// we never scroll the line ends
}

- (void) moveToX:(int)xp Y:(int)yp {
	cx = xp;
	cy = yp;
	if (cx > maxX-1) { cx = maxX-1;};
	if (cx < 0) {cx = 0;};
	if (cy < 0) {cx = 0;};
	if (cy > maxY-1) {cy = maxY-1;};
}

- (void) home {
	[self clearScreen];
	[self moveToX:0 Y:0];
}

- (void) cursorX : (int *) x Y : (int *) y {
	*x = cx;
	*y = cy;
}

- (void) writeChar:(char)theChar {
	char *currline = [self currline]; // get adrs of first char in curent line
	currline[cx] = theChar;
	cx = cx + 1;
	if (cx > maxX-1) {cx = maxX-1;}; 

}

- (void) writeString:(NSString *)aString {
	// writes at current x, y
	// calculate number of chars to write
	int len = [aString length];
	if (len > maxX-1) {len = maxX;};
	if ((len + cx) > maxX) { 
		int toomuch = maxX - (len + cx);
		len = len - toomuch;
	};
	
	// ok, get the unicode string converted to c
	char *theStr = nil;
	theStr = [aString cStringUsingEncoding:NSMacOSRomanStringEncoding];
	int index = 0;
	char *currline = [self currline];

	while (index < len) {
		currline[cx] = theStr[index];
		cx = cx + 1;
		if (cx > maxX-1) {cx = maxX-1;}; 
		index = index + 1;
	}
	if (cx > maxX-1) {cx = maxX-1;};
	
	if ([self autosync]) {[self updateLCD];}; 
	
}

- (void) writeLnString : (NSString *) aString {
	bool syncSave = [self autosync];
	[self setAutosync:false]; // momentarily switch autosync off so we can use write
	[self writeString:aString];
	[self setAutosync:syncSave];
	
	char *currline = [self currline];
	while (cx < maxX) {
		currline[cx] = 0x20;
		cx = cx + 1;
	};
	cx = 0;
	cy = cy + 1;
	if (cy > maxY-1) {
		cy = maxY-1;
		[self scrollLine];
	}
	
	if ([self autosync]) {[self updateLCD];}; 
}

- (void) logAll {
	NSLog([NSString stringWithCString:l1 encoding:NSMacOSRomanStringEncoding]);
	NSLog([NSString stringWithCString:l2 encoding:NSMacOSRomanStringEncoding]);
	NSLog([NSString stringWithCString:l3 encoding:NSMacOSRomanStringEncoding]);
	NSLog([NSString stringWithCString:l4 encoding:NSMacOSRomanStringEncoding]);
	NSLog(@"-----------");
	
}

- (void) setAutosync:(_Bool)sync {
	useAutosync = sync;
}

- (bool) autosync {
	return useAutosync;
}

- (void) updateLCD {
	// does nothing in virtual class, but is present
	// so we can call it for all writes
	[self logAll]; // we do, however, write to log so you can use a virtual device to inspect
				   // your code
}

// device convenience 
- (bool) backlight {return false;}
- (void) setBacklight : (bool) on { 
	if (on) {
		[self backlightOn];
	} else {
		[self backlightOff];
	};

}

- (bool) button: (int) button {return false;}
- (void) setCursor: (bool) on blinking : (bool) blink {}
- (void) getCursor: (bool *) on blinking : (bool *) blink {*on = false; *blink = false;}
// low-level for compatibility
- (void) backlightOn {}
- (void) backlightOff {}
- (void) clearLCD {}
- (void) writeLine : (char *) text atX : (uint8_t) x Y : (uint8_t) y {}
@end

#pragma mark LCD Display 20 x 4
@implementation TDLCD20x4

-(id) initWithUID: (const char *) uid {
	self = [super initWithUID:uid];
	if (self) {
		// create the bricklet
		lcd_20x4_create(&theLCD, uid); 
		maxX = 20;
		maxY = 4;
	}
	return (self);
}

- (Device *) device { 
	return &theLCD; 
}

- (NSString *) deviceType {
	return @"LCD 20x4 Display";
}

- (void) deviceName : (char [40]) theName { // must be at least 40 chars wide
	uint8_t firmware[3];
	uint8_t binding[3];
	
	lcd_20x4_get_version([self device], theName, firmware, binding);
}

- (void) deviceFirmware : (uint8_t [3]) firmware { 
	char theName[40];
	uint8_t binding[3];
	
	lcd_20x4_get_version([self device], theName, firmware, binding);
}


- (void) deviceBinding : (uint8_t [3]) binding { 
	char theName[40];
	uint8_t firmware[3];
	
	lcd_20x4_get_version([self device], theName, firmware, binding);
}

- (void) registerCallback : (void *) callback forID : (uint8_t) theID {
	lcd_20x4_register_callback([self device], theID, callback);
}

// specific
- (void) updateLCD {
	// and here it is - the mother of all complex routines :)
	// this cryptic code copies four lines of text from the 'pseudo lcd' buffer to the device
	[self writeLine:l1 atX:0 Y:0];
	[self writeLine:l2 atX:0 Y:1];
	[self writeLine:l3 atX:0 Y:2];
	[self writeLine:l4 atX:0 Y:3];
}

// low-level
- (bool) backlight {
	bool bl;
	lcd_20x4_is_backlight_on([self device], &bl);
	return bl;
}

- (bool) button: (int) button {
	bool status;
	lcd_20x4_is_button_pressed([self device], button, &status);
	return status;
}

- (void) setCursor: (bool) on blinking : (bool) blink {
	lcd_20x4_set_config([self device], on, blink);
}

- (void) getCursor: (bool *) on blinking : (bool *) blink {
	lcd_20x4_get_config([self device], on, blink);
}

// low-level for compatibility
- (void) backlightOn {lcd_20x4_backlight_on([self device]);}
- (void) backlightOff {lcd_20x4_backlight_off([self device]);}
- (void) clearLCD {lcd_20x4_clear_display([self device]);}
- (void) writeLine : (char *) text atX : (uint8_t) x Y : (uint8_t) y {
	lcd_20x4_write_line([self device], y, x, text);
}

@end






