//
//  TinkerIO.h
//
//  Created by Christian Franz on 23/11/2012.
//  Copyright 2012 cf/x AG. All rights reserved.
//     http://www.cfxsoftware.com
//
//  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and 
//  associated documentation files (the "Software"), to deal in the Software without restriction, including 
//  without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or 
//  sell copies of the Software, and to permit persons to whom the Software is furnished to do so, 
//  subject to the following conditions:
//
//  The above copyright notice and this permission notice shall be included in all copies or substantial 
//  portions of the Software and its documentation.
//
//  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT 
//  NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
//  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
//  WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
//  SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//

#import <Cocoa/Cocoa.h>
#include "TinkerDevices.h"

/*
 TinkerIO is the Cocoa-Equivalent/Wrapper for IPConnection
 
 TinkerIO keeps a list of all connected devices that you can access as follows
 - devices returns an NSArray of all connected devices
 - deviceWithUID: returns the device with a certain UID or nil if not found
 
 TinkerIO supports Exceprions. If an IP operation fails, you can choose to have it raise an Exception
 To do so, use setUseException:
 If an exception is raised, you can retrieve the object that cuased it from the userInfo dictionary with the key @"Object"
 
 You can always query the result of the last operation by looking at lastOp
 
 TinkerIO can be used in 'manual' mode, where you add devices manually with addDevice: or in 
 'Plug&Play' mode where TinkerIO automatically calls a delegate that must conform to the
 TinkerIOProtocol (i.e. have at least the messages 'deviceAdded:', 'deviceRemoved:' implemented
 
 */ 


@protocol TinkerIOProtocol
// provisional protocol. to be verified with new API
- (void) deviceAdded : (TinkerDevice *) aDevice; // called when a device is detected
- (void) deviceRemoved : (TinkerDevice *) aDevice; // called when a device is removed
- (void) connectFailed : (TinkerDevice *) aDevice; // called when a manual connect attempt fails
- (void) connectSuccessful : (TinkerDevice *) aDevice; // called when a manual connect attempt returns ok
@end


@interface TinkerIO : NSObject {
	BOOL usesExceptions; // if true, operations throw exceptions on errors
						// currently not implemented
	BOOL connectionValid;
	int lastOp; // result of last low-level call
	
	IPConnection ipcon; // this is the C-Bindings struct
	
	NSMutableArray *myDevices; // array that holds all connected devices
	id myDelegate; // the delegate object
}
- (id) initWithHost : (const char *) aHost port : (int) aPort; // initializes for 'manual'
- (id) initWithHost : (const char *) aHost port : (int) aPort delegate : (id) aDelegate; // delegate must	
																						  // conform to TinkerIOProtocol
- (void) beginAutoconfig; // once inited as P&P, call this to start scanning and sending Add/Remove messages
						  // currently not implemented

- (void) setUseExceptions : (BOOL) useIt;
- (bool) useException;
- (IPConnection *) connection;
- (int) connectTo : (TinkerDevice *) aDevice; // same as addDevice. simply a better name
- (int) addDevice : (TinkerDevice *) aDevice; // aDevice must be of type TinkerDevice 
- (NSArray *) devices; // return an array of all connected devices
- (TinkerDevice *) deviceWithUID : (NSString *) uid;

- (void) setDelegate: (id) aDelegate;
- (id) delegate;

@end
