//
//  TinkerIO.m
//
//  Created by Christian Franz on 23/11/2012.
//  Copyright 2012 cf/x AG. All rights reserved.
//     http://www.cfxsoftware.com
//
//  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and 
//  associated documentation files (the "Software"), to deal in the Software without restriction, including 
//  without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or 
//  sell copies of the Software, and to permit persons to whom the Software is furnished to do so, 
//  subject to the following conditions:
//
//  The above copyright notice and this permission notice shall be included in all copies or substantial 
//  portions of the Software and its documentation.
//
//  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT 
//  NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
//  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
//  WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE 
//  SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
//

#import "TinkerIO.h"


@implementation TinkerIO
- (id) initWithHost : (const char *) aHost  port : (int) aPort {
	self = [super init];
	if (self) {
		connectionValid = FALSE;
		lastOp = ipcon_create(&ipcon, aHost, aPort);
		if (lastOp != E_OK) {
			connectionValid = FALSE;
		} else { 
			connectionValid = TRUE;
		};
		
		myDevices = [[NSMutableArray alloc] initWithCapacity:10];
	};
	return (self);
}

// the Plug&Play version. Note that it curently does not work because of limitations
// in the current callback API that cannot pass a context

- (id) initWithHost : (const char *) aHost port : (int) aPort delegate : (id) aDelegate {
	self = [self initWithHost : aHost port : aPort];
	if (self) {
		myDelegate = aDelegate;
		[myDelegate retain];
	};
	
	return self;
}

- (void) beginAutoconfig {
	// we would now call enumerate and generate 'deviceAdded' and 'deviceRemoved' messages
	// whenever we detect a device. Alas, we need a context for this to work to be passed
	// back to the callback. Once this is implemented with API 2.0, we'll be able to do that.
	
	// for now, we just don't do anything
}


- (void) dealloc {
	if (connectionValid) {
		// destruct connection
		ipcon_destroy(&ipcon);
		if ([self delegate]) {[[self delegate] release];}
	}
	
	[super dealloc];
}

- (void) setUseExceptions : (BOOL) useIt {
	usesExceptions = useIt;
}

- (bool) useException {return usesExceptions;};

- (IPConnection *) connection {
	if (connectionValid) {
		return &ipcon; 
	} else {
		return nil;
	};
}

- (int) addDevice:(TinkerDevice *) aDevice {
	// if this is a single object and already added, exit with OK
	if (![aDevice isCompoundDevice] & [[self devices] containsObject:aDevice]) {
		lastOp = E_OK;
		return lastOp;
	};
	
	lastOp = -1;
	if (![aDevice isCompoundDevice]) {
		lastOp = ipcon_add_device([self connection], [aDevice device]);
		if (lastOp == E_OK) {
			[myDevices addObject:aDevice];
			[aDevice setValue:[NSNumber numberWithBool:TRUE] forKey:@"isConnected"];
		
			// notify the delegate
			if ([self delegate]) {
				if ([[self delegate] respondsToSelector:@selector(connectSuccessful:)]) {
					[[self delegate] connectSuccessful:aDevice];
				};
			};
		
		} else {
			// raise ecception if applicable
			if ([self useException]) {
				NSException *theException = [[NSException alloc] initWithName:@"TinkerIO" 
																	   reason:[NSString stringWithFormat:@"addDevice:/connectTo: failed with %d", lastOp]
																	 userInfo:[NSDictionary dictionaryWithObject:aDevice forKey:@"Object"]];
				[theException raise];
			};
		
			[aDevice setValue:[NSNumber numberWithBool:FALSE] forKey:@"isConnected"];

			// notify the delegate
			if ([self delegate]) {
				if ([[self delegate] respondsToSelector:@selector(connectFailed:)]) {
					[[self delegate] connectFailed:aDevice];
				};
			};
				 
		}
	} else {
		// we have a compound device. simply recurse
		TinkerDevice *cDevice = nil;
		NSArray *deviceList = [aDevice compoundDevices];
		for (cDevice in deviceList) {
			// add this component if it wasn't connected already
			if (![[self devices] containsObject:cDevice]) {[self addDevice:cDevice];}
		}
	
	}
	
	return lastOp;
}

- (int) connectTo:(TinkerDevice *)aDevice {
	return [self addDevice:aDevice];
}

- (int) removeDevice: (TinkerDevice *) aDevice {
	int theIndex = [myDevices indexOfObject:aDevice];
	if (theIndex >= 0) {
		[myDevices removeObject : aDevice];
		return 0;
	} else {
		return -1;
	};

}

- (NSArray *) devices {
	return myDevices;
}

- (TinkerDevice *) deviceWithUID : (NSString *) uid {	
	TinkerDevice *aDevice;
	NSArray *allDevices = [self devices];
	// some fast enumeration goodness
	for (aDevice in allDevices) {
		if ( [[aDevice deviceName] compare:uid ] == NSOrderedSame) {
			return aDevice;
		}		
	}
	return nil;
}

- (void) setDelegate:(id)aDelegate {
	[aDelegate retain];
	if ([self delegate]) { [[self delegate] release];}
	myDelegate = aDelegate;
}

- (id) delegate {return myDelegate;}
@end
